from ragflow_sdk import RAGFlow
class Client:
    def __init__(self, api_key:str, base_url ="http://localhost:9380/"):
        self.client = RAGFlow(api_key=api_key, base_url=base_url)

    @property
    def datasets(self):
        return self.client.list_datasets()


import time

import requests
from davidkhala.utils.http_request import default_on_response
from requests import Response

from davidkhala.ai.api import API
from davidkhala.ai.model.chat import CompareChatAware


class OpenRouter(API, CompareChatAware):

    @property
    def free_models(self) -> list[str]:
        l = list(
            map(lambda model: model['id'],
                filter(lambda model: model['id'].endswith(':free'), self.list_models())
                )
        )
        l.append('openrouter/free')
        return l

    @staticmethod
    def on_response(response: requests.Response):
        r = default_on_response(response)
        # openrouter special error on response.ok
        err = r.get('error')
        if err:
            derived_response = Response()
            derived_response.status_code = err['code']
            derived_response.reason = err['message']
            derived_response.metadata = err.get("metadata")

            derived_response.raise_for_status()
        return r

    def __init__(self, api_key: str, **kwargs):
        super().__init__(api_key, 'https://openrouter.ai/api')

        if 'leaderboard' in kwargs and type(kwargs['leaderboard']) is dict:
            self.options["headers"]["HTTP-Referer"] = kwargs['leaderboard'][
                'url']  # Site URL for rankings on openrouter.ai.
            self.options["headers"]["X-Title"] = kwargs['leaderboard'][
                'name']  # Site title for rankings on openrouter.ai.
        self.on_response = OpenRouter.on_response
        self.retry = True

    def request(self, url, method: str, params=None, data=None, json=None) -> dict:
        try:
            return super().request(url, method, params, data, json)
        except requests.HTTPError as e:
            if e.response.status_code == 429 and self.retry:  # 429: You are being rate limited
                time.sleep(1)
                return self.request(url, method, params, data, json)
            else:
                raise

    def as_chat(self, *models: str, sys_prompt: str = None):
        CompareChatAware.as_chat(self, *models, sys_prompt=sys_prompt)

    def chat(self, *user_prompt: str, **kwargs):
        if self._models:
            kwargs["models"] = self._models
        else:
            kwargs["model"] = self.model

        r = super().chat(*user_prompt, **kwargs)

        if self._models:
            assert r['model'] in self._models
        return r



from __future__ import annotations

from typing import List, Any, TypedDict

from pydantic import BaseModel

from davidkhala.ai.agent.dify.model.knowledge import Document as DocumentBase


class RerankingModel(BaseModel):
    reranking_provider_name: str | None
    reranking_model_name: str | None


class KeywordSetting(BaseModel):
    keyword_weight: float


class VectorSetting(BaseModel):
    vector_weight: float
    embedding_model_name: str
    embedding_provider_name: str


class Weights(BaseModel):
    weight_type: str | None
    keyword_setting: KeywordSetting | None
    vector_setting: VectorSetting | None


class RetrievalModelDict(BaseModel):
    search_method: str
    reranking_enable: bool
    reranking_mode: str | None
    reranking_model: RerankingModel | None
    weights: Weights | None
    top_k: int | None
    score_threshold_enabled: bool
    score_threshold: float


class ExternalKnowledgeInfo(BaseModel):
    external_knowledge_id: str | None
    external_knowledge_api_id: str | None
    external_knowledge_api_name: str | None
    external_knowledge_api_endpoint: str | None


class ExternalRetrievalModel(BaseModel):
    top_k: int
    score_threshold: float
    score_threshold_enabled: bool


class IconInfo(BaseModel):
    icon_type: str
    icon: str | None
    icon_background: str | None
    icon_url: str | None


class DatasetModel(BaseModel):
    id: str
    name: str
    description: str | None = None
    provider: str
    permission: str
    data_source_type: str | None = None
    indexing_technique: str | None = None
    app_count: int
    document_count: int
    word_count: int
    created_by: str | None = None
    author_name: str | None = None
    created_at: int | None = None
    updated_by: str | None = None
    updated_at: int | None = None
    embedding_model: str | None = None
    embedding_model_provider: str | None = None
    embedding_available: bool
    retrieval_model_dict: RetrievalModelDict | None = None
    tags: List[Any] = []
    doc_form: str | None = None
    external_knowledge_info: ExternalKnowledgeInfo | None = None
    external_retrieval_model: ExternalRetrievalModel | None = None
    doc_metadata: List[Any] = []
    built_in_field_enabled: bool
    pipeline_id: str | None = None
    runtime_mode: str | None = None
    chunk_structure: str | None = None
    icon_info: IconInfo | None = None
    is_published: bool
    total_documents: int | None = None
    total_available_documents: int | None = None
    enable_api: bool
    is_multimodal: bool


class DocumentModel(DocumentBase):
    data_source_info: dict[str, str]
    data_source_detail_dict: dict[str, dict]
    dataset_process_rule_id: str
    created_from: str
    created_by: str
    created_at: int
    tokens: int
    archived: bool
    display_status: str
    word_count: int
    hit_count: int
    doc_form: str
    doc_metadata: dict
    disabled_at: int
    disabled_by: str


class ChunkDict(TypedDict):
    id: str
    position: int
    document_id: str
    content: str
    sign_content: str  # trimmed version of content
    answer: str | None  # only used in QA chunk
    word_count: int
    tokens: int
    keywords: list[str] | None
    index_node_id: str  # chunk 在向量索引中的节点 ID
    index_node_hash: str  # hash of sign_content
    hit_count: int
    enabled: bool
    status: str  # 'completed'
    created_at: int  # timestamp
    updated_at: int  # timestamp
    completed_at: int  # timestamp
    created_by: str  # user id
    child_chunks: list
    error: Any | None
    stopped_at: int | None  # timestamp
    disabled_at: int | None  # timestamp



from mistralai import Mistral


class Client:

    def __init__(self, api_key: str):
        self.api_key = api_key
        self.client = Mistral(api_key=api_key)

    def __enter__(self):
        self.client.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.client.__exit__(exc_type, exc_val, exc_tb)



# https://github.com/mistralai/client-python

from mistralai import ResponseFormat

from davidkhala.ai.mistral import Client as MistralClient
from davidkhala.ai.model.embed import EmbeddingAware
from davidkhala.ai.model.chat import on_response, ChatAware


class Client(ChatAware, EmbeddingAware, MistralClient):
    def __init__(self, api_key: str):
        ChatAware.__init__(self)
        MistralClient.__init__(self, api_key)


    def as_chat(self, model="mistral-large-latest", sys_prompt: str = None):
        super().as_chat(model, sys_prompt)

    def as_embeddings(self, model="mistral-embed"):
        super().as_embeddings(model)

    def chat(self, *user_prompt, **kwargs):
        response = self.client.chat.complete(
            model=self.model,
            messages=self.messages_from(*user_prompt), stream=False, response_format=ResponseFormat(type='text'),
            n=self.n,
        )

        return on_response(response, self.n)

    def encode(self, *_input: str) -> list[list[float]]:
        res = self.client.embeddings.create(
            model=self.model,
            inputs=_input,
        )
        return [d.embedding for d in res.data]

    @property
    def models(self) -> list[str]:
        return [_.id for _ in self.client.models.list().data]



from typing import Literal, Union

from mistralai import Agent, ToolExecutionEntry, FunctionCallEntry, MessageOutputEntry, AgentHandoffEntry

from davidkhala.ai.mistral import Client as MistralClient
from davidkhala.ai.model.chat import messages_from


class Agents(MistralClient):
    def __init__(self, api_key):
        super().__init__(api_key)
        self.instructions: str | None = None
        self.model = None

    def as_chat(self, model="mistral-large-latest", sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.instructions = sys_prompt

    def create(self, name,
               *,
               web_search: Literal["web_search", "web_search_premium"] = None
               ) -> Agent:
        """
        :param name:
        :param web_search:
            "web_search_premium": beyond search engine, add news provider as source
        :return:
        """
        tools = []
        if web_search:
            tools.append({"type": web_search})
        agent = self.client.beta.agents.create(
            model=self.model,
            name=name,
            tools=tools,
            instructions=self.instructions
        )
        return agent

    def chat(self, agent_id: str, *user_prompt: str) -> tuple[
        list[Union[ToolExecutionEntry, FunctionCallEntry, MessageOutputEntry, AgentHandoffEntry]],
        str
    ]:
        response = self.client.beta.conversations.start(
            agent_id=agent_id,
            inputs=messages_from(*user_prompt)
        )

        return response.outputs, response.conversation_id



import voyageai

from davidkhala.ai.model import SDKProtocol
from davidkhala.ai.model.embed import EmbeddingAware


class Client(EmbeddingAware, SDKProtocol):
    def __init__(self, api_key):
        self.client = voyageai.Client(
            api_key=api_key,  # Or use VOYAGE_API_KEY environment variable
        )

    def as_embeddings(self, model: str = 'voyage-4'):
        """
        :param model: see in https://www.mongodb.com/docs/voyageai/models/#choosing-a-model
        """
        super().as_embeddings(model)

    def encode(self, *_input: str) -> list[list[float]]:
        result = self.client.embed(
            texts=list(_input),
            model=self.model
        )
        return result.embeddings



from google.adk.agents import Agent

def with_opik(agent:Agent, **options):
    from opik.integrations.adk import OpikTracer

    _ = OpikTracer(
        name=agent.name,
        metadata={
            "model": agent.model,
            "framework": "google-adk",
        },
        **options
    )
    agent.before_agent_callback = _.before_agent_callback
    agent.after_agent_callback = _.after_agent_callback
    agent.before_model_callback = _.before_model_callback
    agent.after_model_callback = _.after_model_callback
    agent.before_tool_callback = _.before_tool_callback
    agent.after_tool_callback = _.after_tool_callback




from davidkhala.ai.api import API


class SiliconFlow(API):
    @property
    def free_models(self) -> list[str]:
        """
        Cannot be lively fetched by list_models
        """
        return [
            # chat section
            'THUDM/GLM-4.1V-9B-Thinking'
            'THUDM/GLM-Z1-9B-0414'
            'THUDM/GLM-4-9B-0414'
            'THUDM/glm-4-9b-chat'
            'Qwen/Qwen3-8B'
            'Qwen/Qwen2.5-7B-Instruct'
            'Qwen/Qwen2.5-Coder-7B-Instruct'
            'internlm/internlm2_5-7b-chat'
            'deepseek-ai/DeepSeek-R1-0528-Qwen3-8B',
            'deepseek-ai/DeepSeek-R1-Distill-Qwen-7B',
            # embedding and reranker
            'BAAI/bge-m3'
            'BAAI/bge-reranker-v2-m3'
            'BAAI/bge-large-zh-v1.5'
            'BAAI/bge-large-en-v1.5'
            'netease-youdao/bce-reranker-base_v1'
            'netease-youdao/bce-embedding-base_v1'
            # Audio
            'FunAudioLLM/SenseVoiceSmall'
            # image
            'Kwai-Kolors/Kolors'
        ]

    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.siliconflow.cn')

    def chat(self, *user_prompt: str):
        return super().chat(*user_prompt, model=self.model, timeout=50)

    def encode(self, *_input: str) -> list[list[float]]:
        json = {
            'input': _input,
            'model': self.model
        }
        response = self.request(f"{self.base_url}/embeddings", "POST", json=json)
        return [_['embedding'] for _ in response['data']]

    def which(self, query: str, documents: list[str], **kwargs)->tuple[str,int]:
        json = {
            'model': self.model,
            'query': query,
            'documents': documents,
            **kwargs
        }
        response = self.request(f"{self.base_url}/rerank", "POST", json=json)
        most_relevant_index = max(response['results'], key=lambda x: x['relevance_score'])['index']

        return documents[most_relevant_index], most_relevant_index




from enum import Enum


class IndexingStatus(str, Enum):
    WAITING = "waiting"
    PARSING = "parsing"
    SPLITTING = 'splitting'
    INDEXING = "indexing"
    COMPLETED = "completed"
    FAILED = "error"



from openai import OpenAI
from davidkhala.ai.openai import Client as BaseClient


class Client(BaseClient):
    def __init__(self, host: str, token: str):
        super().__init__(OpenAI(
            base_url=f"https://{host}/serving-endpoints",
            api_key=token
        ))

    def chat(self, *user_prompt, **kwargs):
        """Databricks always reasoning"""
        rs = super().chat(*user_prompt, **kwargs)
        for r in rs:
            assert len(r) == 2
            assert r[0]['type'] == 'reasoning'
            for s in r[0]['summary']:
                assert s['type'] =='summary_text'
                yield s['text']
            assert r[1]['type'] == 'text'
            yield r[1]['text']




from davidkhala.ai.agent.dify.ops.db import DB
from davidkhala.ai.agent.dify.ops.db.orm import AppModelConfig
from sqlalchemy.orm import Session
from sqlalchemy import desc

class Studio(DB):
    user_feedbacks_sql = """SELECT mf.conversation_id,
                        mf.content,
                        m.query,
                        m.answer
                 FROM message_feedbacks mf
                          LEFT JOIN messages m ON mf.message_id = m.id
                 WHERE mf.from_source = 'user'"""

    @property
    def apps(self): return self.get_dict("select id, name, mode from apps where status = 'normal'")

    def app_config(self, app_id) -> AppModelConfig | None:
        with Session(self.client) as session:
            return (
                session.query(AppModelConfig)
                .filter(AppModelConfig.app_id == app_id)
                .order_by(desc(AppModelConfig.created_at))
                .first()
            )

    def update_app_config(self, record: AppModelConfig, refresh: bool = False) -> AppModelConfig | None:
        with Session(self.client) as session:
            session.add(record)
            session.commit()
            if refresh:
                session.refresh(record)
                return record
            return None


from pathlib import Path

from mistralai import UploadFileOut, FileSchema, ListFilesOut

from davidkhala.ai.mistral import Client as MistralClient


class Client(MistralClient):
    def upload(self, path: Path, file_name=None) -> str:
        """
        specific schema is required
        - for [Text & Vision Fine-tuning](https://docs.mistral.ai/capabilities/finetuning/text_vision_finetuning)
        - for [Classifier Factory](https://docs.mistral.ai/capabilities/finetuning/classifier_factory)
        :param path:
        :param file_name:
        :return:
        """
        if not file_name:
            file_name = path.name
        assert file_name.endswith(".jsonl"), "Data must be stored in JSON Lines (.jsonl) files"
        with open(path, "rb") as content:
            res: UploadFileOut = self.client.files.upload(file={
                "file_name": file_name,
                "content": content
            })
        return res.id

    def paginate_files(self, page=0, size=100) -> ListFilesOut:
        return self.client.files.list(page=page, page_size=size)

    def ls(self, page_size=100) -> list[FileSchema]:
        has_next = True
        result = []
        while has_next:
            page = self.paginate_files(size=page_size)
            has_next = page.total == page_size
            result.extend(page.data)
        return result



from typing import AsyncIterator

from youdotcom import You, models
from youdotcom.models import Web
from youdotcom.types.typesafe_models import AgentType, get_text_tokens, Format
from youdotcom.utils import eventstreaming


class Client:
    def __init__(self, api_key: str):
        self.client = You(api_key_auth=api_key)

    def chat(self, user_prompt: str, *, tools: list[models.Tool] = None) -> str:
        res = self.client.agents.runs.create(
            agent=AgentType.ADVANCED if tools else AgentType.EXPRESS,
            input=user_prompt,
            stream=False,
            tools=tools,
        )

        return "".join(get_text_tokens(res))

    def scrape(self, *urls: str) -> list[str]:
        """
        :return the content of the web pages as Markdown format (incl. metadata)
        """
        res = self.client.contents.generate(
            urls=list(urls),
            format_=Format.MARKDOWN,
        )
        return [_.markdown for _ in res]

    def search(self, query: str) -> list[Web]:
        res = self.client.search.unified(
            query=query
        )
        return res.results.web

    async def async_chat(self, user_prompt: str) -> AsyncIterator[str]:
        res: eventstreaming.EventStreamAsync[models.Data] = await self.client.agents.runs.create_async(
            agent=AgentType.EXPRESS,
            input=user_prompt,
            stream=True,
        )

        async for event in res:
            if event.type == 'response.output_text.delta':
                yield event.response.delta

    def __enter__(self):
        self.client.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



from openrouter import OpenRouter
from openrouter.components import Model
from openrouter.errors import UnauthorizedResponseError
from openrouter.operations import ListData

from davidkhala.ai.model import Connectable, SDKProtocol
from davidkhala.ai.model.chat import CompareChatAware, on_response
from davidkhala.ai.model.garden import GardenAlike


class Client(CompareChatAware, Connectable, SDKProtocol, GardenAlike):
    def __init__(self, api_key: str):
        super().__init__()
        self.client: OpenRouter = OpenRouter(api_key)

    def chat(self, *user_prompt, **kwargs):
        r = self.client.chat.send(
            model=self.model,
            models=self._models,
            messages=self.messages_from(*user_prompt)
            # openrouter has no n
        )
        return on_response(r, None)

    def connect(self):
        try:
            _ = self.models
            return True
        except UnauthorizedResponseError:
            return False

    def list_models(self) -> list[Model]:
        return self.client.models.list().data

    @property
    def models(self) -> list[str]:
        return [m.id for m in self.list_models()]


class Admin:
    def __init__(self, provisioning_key: str):
        self.provisioning_key = provisioning_key
        self.client = OpenRouter(provisioning_key)

    @property
    def keys(self) -> list[ListData]:
        return self.client.api_keys.list().data



model_provider = [
    'langgenius/siliconflow',
    'langgenius/azure_openai',
    'langgenius/tongyi',
    'langgenius/jina',
    'langgenius/openrouter',
    'langgenius/deepseek',
]

class Knowledge:
    data_source = [
        'langgenius/firecrawl_datasource',
    ]
    chunk = [
        'langgenius/parentchild_chunker',
        'langgenius/general_chunker',
    ]
    api = [
        'abesticode/knowledge_pro',
    ]

class Node:
    format = [
        'langgenius/json_process',
        'langgenius/dify_extractor',
    ]
    agent = [
        'langgenius/agent',
    ]
    data = [
        'langgenius/chart',
        'junjiem/db_query',
        'junjiem/db_query_pre_auth',
    ]
    web = [
        'langgenius/searxng',
        'langgenius/firecrawl'
    ]




from davidkhala.ai.model import ModelAware


class EmbeddingAware(ModelAware):
    def as_embeddings(self, model: str):
        self.model = model

    def encode(self, *_input: str) -> list[list[float]]: ...



import json
from enum import Enum
from typing import Any

from pydantic import BaseModel
from sqlalchemy import Column, String, Text, JSON, TIMESTAMP, func
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import declarative_base

from davidkhala.ai.agent.dify.model.workflow import Node, Position

Base = declarative_base()


class DifyBase(Base):
    __abstract__ = True  # keyword for SQLAlchemy
    id = Column(UUID(as_uuid=True), primary_key=True, server_default=func.uuid_generate_v4())


class AppModelConfig(DifyBase):
    __tablename__ = "app_model_configs"
    __table_args__ = {"schema": "public"}

    app_id = Column(UUID(as_uuid=True), nullable=False)

    provider = Column(String(255))
    model_id = Column(String(255))
    configs = Column(JSON)

    created_at = Column(TIMESTAMP, nullable=False, server_default=func.current_timestamp())
    updated_at = Column(TIMESTAMP, nullable=False, server_default=func.current_timestamp())

    opening_statement = Column(Text)
    suggested_questions = Column(Text)
    suggested_questions_after_answer = Column(Text)
    more_like_this = Column(Text)
    model = Column(Text)
    user_input_form = Column(Text)
    pre_prompt = Column(Text)
    agent_mode = Column(Text)
    speech_to_text = Column(Text)
    sensitive_word_avoidance = Column(Text)
    retriever_resource = Column(Text)

    dataset_query_variable = Column(String(255))
    prompt_type = Column(String(255), nullable=False, server_default="simple")

    chat_prompt_config = Column(Text)
    completion_prompt_config = Column(Text)
    dataset_configs = Column(Text)
    external_data_tools = Column(Text)
    file_upload = Column(Text)
    text_to_speech = Column(Text)

    created_by = Column(UUID(as_uuid=True))
    updated_by = Column(UUID(as_uuid=True))

    def __repr__(self):
        return f"<AppModelConfig(id={self.id}, app_id={self.app_id}, provider={self.provider}, model_id={self.model_id})>"


class Graph:
    @staticmethod
    def convert(*records: list[dict]):
        return [{**record, "graph": Graph(**json.loads(record["graph"]))} for record in records]



from typing import Protocol, Any, Iterable, TypedDict

from davidkhala.ai.model import ModelAware


class MessageProtocol(Protocol):
    content: str | Any


class ChoiceProtocol(Protocol):
    message: MessageProtocol


class ChoicesAware(Protocol):
    choices: list[ChoiceProtocol]


class ImagePromptDict(TypedDict):
    text: str
    image_url: list[str]


def on_response(response: ChoicesAware, n: int | None):
    contents = [choice.message.content for choice in response.choices]
    if n:
        assert len(contents) == n, f"expected {n} choices, but got {len(contents)}"
    return contents


class MessageDict(TypedDict):
    content: str | list | None
    role: str


def messages_from(*user_prompt: str | ImagePromptDict) -> Iterable[MessageDict]:
    for _ in user_prompt:
        message = MessageDict(role='user', content=None)
        if type(_) == str:
            message['content'] = _
        elif type(_) == dict:
            message['content'] = [{"type": "text", "text": _['text']}]
            message['content'].extend({"type": "image_url", "image_url": {"url": i}} for i in _['image_url'])
        yield message


class ChatAware(ModelAware):
    def __init__(self):
        super().__init__()
        self.messages: list[Any | MessageDict] = []
        self.n: int = 1

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.messages = [MessageDict(role='system', content=sys_prompt)]

    def chat(self, *user_prompt, **kwargs): ...

    def messages_from(self, *user_prompt) -> list[MessageDict]:
        messages = list(self.messages)
        messages.extend(messages_from(*user_prompt))
        return messages


class CompareChatAware(ChatAware):
    def __init__(self):
        super().__init__()
        self._models = None

    def as_chat(self, *models: str, sys_prompt: str = None):
        if len(models) > 1:
            self._models = models
            super().as_chat(None, sys_prompt)
        elif len(models) == 1:
            super().as_chat(models[0], sys_prompt)



from time import sleep

from davidkhala.ai.agent.dify.console.session import ConsoleDerived
from davidkhala.ai.agent.dify.const import IndexingStatus
from davidkhala.ai.agent.dify.interface import IndexingError


class Operation(ConsoleDerived):
    def website_sync(self, dataset: str, document: str, *, wait_until=True):
        """
        cannot be used towards a pipeline dataset. Otherwise, you will see error "no website import info found"
        """
        doc_url = f"{self.base_url}/datasets/{dataset}/documents/{document}"

        r = self.request(f"{doc_url}/website-sync", "GET")
        assert r == {"result": "success"}
        if wait_until:
            return self.wait_until(dataset, document)
        return None

    def retry(self, dataset: str, *documents: str, wait_until=True):
        """
        It cannot trigger rerun on success documents
        """
        url = f"{self.base_url}/datasets/{dataset}/retry"
        self.request(url, "POST", json={
            'document_ids': documents,
        })
        # response status code will be 204
        if wait_until:
            return [self.wait_until(dataset, document) for document in documents]
        return None

    def rerun(self, dataset: str, *documents: str):
        for document in documents:
            try:
                self.website_sync(dataset, document)
                assert False, "expect IndexingError"
            except IndexingError:
                pass
        return self.retry(dataset, *documents)

    def wait_until(self, dataset: str, document: str, *,
                   expect_status=None,
                   from_status=None,
                   interval=1
                   ):
        if not expect_status:
            expect_status = [IndexingStatus.FAILED, IndexingStatus.COMPLETED]
        url = f"{self.base_url}/datasets/{dataset}/documents/{document}/indexing-status"
        if from_status is None:
            from_status = [IndexingStatus.WAITING, IndexingStatus.PARSING]
        r = self.request(url, "GET")
        status = r['indexing_status']
        assert status in from_status, f"current status: {status}, expect: {from_status}"
        while status not in expect_status:
            sleep(interval)
            r = self.request(url, "GET")
            status = r['indexing_status']
        if status == IndexingStatus.FAILED: raise IndexingError(r['error'])
        return r



from enum import Enum

from davidkhala.utils.http_request import Request


class API(Request):

    def __init__(self, base_url='http://localhost'):
        """
        :param base_url: "{protocol}://{host}". For Dify cloud, it is 'https://cloud.dify.ai'
        """
        super().__init__()
        self.base_url = f"{base_url}/console/api"
        self.open()




import base64
import json
from pathlib import Path

from davidkhala.ml.ocr.interface import FieldProperties as BaseFieldProperties
from mistralai import ImageURLChunk, ResponseFormat, JSONSchema

from davidkhala.ai.mistral import Client as MistralClient


class FieldProperties(BaseFieldProperties):
    description: str = ""


class Client(MistralClient):
    def process(self, file: Path, schema: dict[str, FieldProperties] = None) -> list[dict]|dict:
        """
        Allowed formats are JPEG, PNG, WEBP, GIF, MPO, HEIF, AVIF, BMP, TIFF
        """
        with open(file, "rb") as f:
            content = base64.b64encode(f.read()).decode('utf-8')
        options = {}
        if schema:
            required = [k for k, _ in schema.items() if _.required]
            properties = {k: {'type': v.type, 'description': v.description} for k, v in schema.items()}
            options['document_annotation_format'] = ResponseFormat(
                type='json_schema',
                json_schema=JSONSchema(
                    name='-',
                    schema_definition={
                        "required": required,
                        "properties": properties
                    },
                    strict=True
                )
            )

        ocr_response = self.client.ocr.process(
            model="mistral-ocr-latest",
            document=ImageURLChunk(image_url=f"data:image/jpeg;base64,{content}"),
            include_image_base64=True,
            **options,
        )
        if schema:
            return json.loads(ocr_response.document_annotation)
        return [{'markdown': page.markdown, 'images': page.images} for page in ocr_response.pages]



import runpy

from google.genai import Client


def with_opik(client: Client) -> Client:
    from opik.integrations.genai import track_genai
    runpy.run_path('../opik.py')
    return track_genai(client)



from pydantic import BaseModel

from davidkhala.ai.agent.dify.plugins import DataSourceTypeAware


class DataSourceInfo(BaseModel):
    source_url: str
    content: str
    title: str
    description: str


class DataSourceOutput(DataSourceTypeAware, DataSourceInfo):
    datasource_type: str = "website_crawl"


class CredentialAware(BaseModel):
    credential_id: str | None


class Console(DataSourceOutput, CredentialAware):
    pass



from typing import Any

from davidkhala.utils.http_request.stream import Request as StreamRequest, as_sse
from pydantic import BaseModel, Field

from davidkhala.ai.agent.dify.console.session import ConsoleDerived
from davidkhala.ai.agent.dify.model import User
from davidkhala.ai.agent.dify.model.knowledge import Dataset, Document
from davidkhala.ai.agent.dify.model.workflow import NodeProtocol, Graph


class RAGPipelineVariable(BaseModel):
    label: str
    variable: str
    type: str
    belong_to_node_id: str
    max_length: int | None = None
    required: bool = False
    unit: str | None = None
    default_value: Any | None = None
    options: list[Any] = Field(default_factory=list)
    placeholder: str | None = None
    tooltips: str | None = None
    allowed_file_types: str | None = None
    allow_file_extension: str | None = None
    allow_file_upload_methods: str | None = None


class PipelineModel(BaseModel):
    id: str
    graph: Graph
    features: dict[str, Any] = Field(default_factory=dict)
    hash: str
    version: str
    marked_name: str = ""
    marked_comment: str = ""
    created_by: User
    created_at: int
    updated_by: User | None = None
    updated_at: int
    tool_published: bool = False
    environment_variables: list[dict[str, Any]]
    conversation_variables: list[dict[str, Any]]
    rag_pipeline_variables: list[RAGPipelineVariable]


class DatasetResult(Dataset):
    chunk_structure: str


class RunResult(BaseModel):
    batch: str
    dataset: DatasetResult
    documents: list[Document]


class Pipeline(ConsoleDerived):

    def async_run(self, pipeline: str, node: NodeProtocol, inputs: dict, datasource_info_list: list[dict]) -> RunResult:
        """Ingest new document"""
        url = f"{self.base_url}/rag/pipelines/{pipeline}/workflows/published/run"
        r = self.request(url, "POST", json={
            'inputs': inputs,
            'start_node_id': node.id,
            'is_preview': False,
            'response_mode': "blocking",
            "datasource_info_list": datasource_info_list,
            'datasource_type': node.datasource_type
        })
        return RunResult.model_validate(r)

    def get(self, pipeline: str):
        url = f"{self.base_url}/rag/pipelines/{pipeline}/workflows/publish"
        r = self.request(url, "GET")
        return PipelineModel.model_validate(r)


class Datasource(ConsoleDerived):
    class FirecrawlOutput(BaseModel):
        source_url: str
        description: str
        title: str
        credential_id: str
        content: str

    def run_firecrawl(self, pipeline: str, node: NodeProtocol,
                      *,
                      inputs: dict,
                      credential_id: str
                      ):

        url = f"{self.base_url}/rag/pipelines/{pipeline}/workflows/published/datasource/nodes/{node.id}/run"

        stream_request = StreamRequest(self)
        response = stream_request.request(url, 'POST', json={
            'inputs': inputs,
            'datasource_type': node.datasource_type,
            'credential_id': credential_id,
            "response_mode": "streaming"
        })

        for data in as_sse(response):
            event = data['event']
            if event == 'datasource_completed':
                return data['data']
            else:
                assert event == 'datasource_processing'
                print(data)
        return None

    def upload(self):
        "http://localhost/console/api/files/upload?source=datasets"
        # TODO
        "form data"
        {
            "file": "body"
        }
        r = {
            "id": "3898db5b-eb72-4f11-b507-628ad5d28887",
            "name": "Professional Diploma Meister Power Electrical Engineering - Technological and Higher Education Institute of Hong Kong.html",
            "size": 254362,
            "extension": "html",
            "mime_type": "text\/html",
            "created_by": "dbd0b38b-5ef1-4123-8c3f-0c82eb1feacd",
            "created_at": 1764943811,
            "source_url": "\/files\/3898db5b-eb72-4f11-b507-628ad5d28887\/file-preview?timestamp=1764943811&nonce=43b0ff5a13372415be79de4cc7ef398c&sign=7OJ2wiVYc4tygl7yvM1sPn7s0WXDlhHxgX76bsGTD94%3D"
        }



from enum import Enum
from typing import Protocol, Literal, Any, Optional

from pydantic import BaseModel, Field




class NodeProtocol(Protocol):
    id:str
    datasource_type: str


class Position(BaseModel):
    x: float
    y: float
class Viewport(Position):
    zoom: float

class JsonData(BaseModel):
    data: list


class NodeOutput(BaseModel):
    """Schema for Output of a Dify node"""
    text: str
    files: list
    json_: list[JsonData] = Field(alias="json") # avoid conflict with .json()




class NodeData(BaseModel):
    class Type(str, Enum):
        SOURCE = 'datasource'
        CHUNKER = 'knowledge-index'
        TOOL = 'tool'

    type: Type | str  # not limit to built-in types
    title: str | None = None
    selected: bool

    # datasource
    datasource_parameters: dict[str, Any] | None = None
    datasource_configurations: dict[str, Any] | None = None
    plugin_id: str | None = None
    provider_type: str | None = None
    provider_name: str | None = None
    datasource_name: str | None = None
    datasource_label: str | None = None
    plugin_unique_identifier: str | None = None

    # tool
    tool_parameters: dict[str, Any] | None = None
    tool_configurations: dict[str, Any] | None = None
    tool_node_version: str | None = None
    provider_id: str | None = None
    provider_icon: str | None = None
    tool_name: str | None = None
    tool_label: str | None = None
    tool_description: str | None = None
    is_team_authorization: bool | None = None
    paramSchemas: list[Any] | None = None
    params: dict[str, Any] | None = None

    # knowledge index
    index_chunk_variable_selector: list[str] | None = None
    keyword_number: int | None = None
    retrieval_model: dict[str, Any] | None = None
    chunk_structure: str | None = None
    indexing_technique: str | None = None
    embedding_model: str | None = None
    embedding_model_provider: str | None = None

class Node(BaseModel):
    @property
    def datasource_type(self): return self.data.provider_type
    id: str
    type: Literal['custom']
    data: NodeData
    position: Position
    targetPosition: str | None = None
    sourcePosition: str | None = None
    positionAbsolute: Position | None = None
    width: float | None = None
    height: float | None = None
    selected: bool | None = False

class EdgeData(BaseModel):
    sourceType: str | None = None
    targetType: str | None = None
    isInIteration: bool | None = False
    isInLoop: bool | None = False
class Edge(BaseModel):
    id: str
    type: str
    source: str
    target: str
    sourceHandle: str | None = None
    targetHandle: str | None = None
    data: EdgeData | None = None
    zIndex: int | None = None


class Graph(BaseModel):
    nodes: list[Node]
    edges: list[Edge]
    viewport: Viewport

    @property
    def datasources(self):
        return [node for node in self.nodes if node.data.type == NodeData.Type.SOURCE]




import base64
import json
from pathlib import Path

from davidkhala.ml.ocr.interface import FieldProperties
from davidkhala.utils.syntax.compat import deprecated
from openai import AzureOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletionUserMessageParam, ChatCompletionContentPartTextParam, ChatCompletionContentPartImageParam
)
from openai.types.chat.chat_completion_content_part_image_param import ImageURL
from openai.types.shared_params import ResponseFormatJSONSchema
from openai.types.shared_params.response_format_json_schema import JSONSchema

from davidkhala.ai.model.chat import on_response
from davidkhala.ai.openai import Client


class AzureHosted(Client):
    def chat(self, *user_prompt, **kwargs):
        if 'web_search_options' in kwargs:
            raise ValueError('Web search options not supported in any models of Azure AI Foundry')
        return super().chat(*user_prompt, **kwargs)


class ModelDeploymentClient(AzureHosted):
    def __init__(self, key, deployment):
        super().__init__(AzureOpenAI(
            api_version="2024-12-01-preview",  # mandatory
            azure_endpoint=f"https://{deployment}.cognitiveservices.azure.com/",
            api_key=key,
        ))

    def process(self, file: Path, schema: dict[str, FieldProperties])-> list[dict]:
        with open(file, "rb") as f:
            content = base64.b64encode(f.read()).decode("utf-8")
        required = [k for k, _ in schema.items() if _.required]
        properties = {k: {'type': v.type} for k, v in schema.items()}

        json_schema = JSONSchema(
            name='-',
            schema={"type": "object",
                    "properties": properties,
                    "required": required,
                    },
        )

        self.messages.append(ChatCompletionUserMessageParam(
            role='user',
            content=[
                ChatCompletionContentPartTextParam(
                    type='text',
                    text="Extract the required fields from this image and return the output strictly following the provided JSON schema."),
                ChatCompletionContentPartImageParam(
                    type="image_url",
                    image_url=ImageURL(
                        url=f"data:image/jpeg;base64,{content}",
                        detail='auto'
                    )
                )
            ]
        ))
        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.messages,
            response_format=ResponseFormatJSONSchema(
                type='json_schema',
                json_schema=json_schema
            ),
            n=self.n,
        )
        return [json.loads(_) for _ in on_response(response, self.n)]

@deprecated("Azure Open AI is deprecated. Please migrate to Microsoft Foundry")
class OpenAIClient(AzureHosted):

    def __init__(self, api_key, project):
        super().__init__(OpenAI(
            base_url=f"https://{project}.openai.azure.com/openai/v1/",
            api_key=api_key,
        ))

    def as_chat(self, model="gpt-oss-120b", sys_prompt: str = None):
        super().as_chat(model, sys_prompt)



import runpy

from openai import OpenAI, AsyncOpenAI
from typing import Union


def bind(instance: Union[OpenAI, AsyncOpenAI]):
    from opik.integrations.openai import track_openai
    runpy.run_path('../opik.py')
    return track_openai(instance)


import datetime

from davidkhala.utils.http_request import Request

from davidkhala.ai.model.chat import ChatAware
from davidkhala.ai.model.garden import GardenAlike


class API(ChatAware, Request, GardenAlike):
    def __init__(self, api_key: str, base_url: str):
        ChatAware.__init__(self)
        Request.__init__(self, {
            "bearer": api_key
        })
        self.base_url = base_url + '/v1'

    def chat(self, *user_prompt: str, **kwargs):
        json = {
            "messages": self.messages_from(*user_prompt),
            **kwargs,
        }

        response = self.request(f"{self.base_url}/chat/completions", "POST", json=json)

        return {
            "data": list(map(lambda x: x['message']['content'], response['choices'])),
            "meta": {
                "usage": response['usage'],
                "created": datetime.datetime.fromtimestamp(response['created'])
            },
            'model': response['model'],
        }

    def list_models(self):
        response = self.request(f"{self.base_url}/models", "GET")
        return response['data']



from __future__ import annotations

from typing import Iterable

import requests

from davidkhala.ai.agent.dify.api import API, Iterator
from davidkhala.ai.agent.dify.api.knowledge.dataset import Dataset
from davidkhala.ai.agent.dify.api.knowledge.model import ChunkDict


class Document(API):
    def __init__(self, d: Dataset.Instance, document_id: str):
        super().__init__(d.api_key, f"{d.base_url}/documents/{document_id}")

    def exist(self):
        try:
            self.get()
            return True
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                raise e

    def get(self):
        return self.request(self.base_url, "GET")

    def paginate_chunks(self, page=1, size=20):
        return self.request(f"{self.base_url}/segments", "GET", params={
            'page': page,
            'limit': size
        })

    def list_chunks(self) -> Iterable[ChunkDict]:
        for chunk_batch in Iterator(self.paginate_chunks, None):
            for chunk in chunk_batch:
                yield chunk

    def delete(self):
        if self.exist():
            self.request(self.base_url, "DELETE")



from time import sleep

from davidkhala.ai.agent.dify.console.session import ConsoleUser, ConsoleDerived


class ConsolePlugin(ConsoleDerived):
    def __init__(self, context: ConsoleUser):
        super().__init__(context)
        self.base_url = f"{context.base_url}/workspaces/current/plugin"

    def upgrade(self, *plugin_names: str) -> list[dict]:
        versions = self.latest_version(*plugin_names)
        self.async_install(*versions)

        current = []
        while len(current) < len(versions):
            current = self.get(*plugin_names)
            sleep(1)
        return current

    def async_install(self, *plugin_versioned_names: str) -> str | None:
        url = f"{self.base_url}/install/marketplace"
        r = self.request(url, method="POST", json={
            'plugin_unique_identifiers': plugin_versioned_names,
        })
        if r['all_installed']:
            # plugins exist, no need to install
            return None

        return r['task_id']

    def plugins(self, *, page=1, size=100):
        url = f"{self.base_url}/list?page={page}&page_size={size}"
        r = self.request(url, method="GET")
        _ = r['plugins']
        assert r['total'] == len(_)
        return _

    def get(self, *plugin_names: str) -> list[dict]:
        """inspect installed plugins"""
        url = f"{self.base_url}/list/installations/ids"
        r = self.request(url, method="POST", json={
            'plugin_ids': plugin_names,
        })
        return r['plugins']

    def latest_version(self, *plugin_names: str) -> dict:
        url = f"{self.base_url}/list/latest-versions"
        r = self.request(url, method="POST", json={
            'plugin_ids': plugin_names,
        })
        return [r['versions'][name]['unique_identifier'] for name in plugin_names]

    def uninstall(self, id: str):
        url = f"{self.base_url}/uninstall"

        r = self.request(url, method="POST", json={
            'plugin_installation_id': id
        })
        assert r['success'] is True

    def uninstall_by(self, *plugin_names: str):
        for name in plugin_names:
            r = self.get(name)
            self.uninstall(r[0]['id'])


class ConsoleTool(ConsoleDerived):
    def __init__(self, context: ConsoleUser):
        super().__init__(context)
        self.base_url = f"{context.base_url}/workspaces/current/tool-provider"

    def credential_id_by(self, name, owner: str, plugin: str):
        for c in self.credentials_of(owner, plugin):
            if c['name'] == name:
                return c['id']
        return None

    def credentials_of(self, owner: str, plugin: str):
        url = f"{self.base_url}/builtin/{owner}/{plugin}/{plugin}/credential/info"
        r = self.request(url, method="GET")
        return r['credentials']



from base64 import b64encode

from requests.cookies import RequestsCookieJar

from davidkhala.ai.agent.dify.console import API


class ConsoleUser(API):
    def login(self, email, password,
              *,
              remember_me=True,
              language="en-US"
              ) -> RequestsCookieJar:
        url = f"{self.base_url}/login"

        r = self.request(url, "POST", json={
            'email': email,
            'password': b64encode(password.encode()).decode(),  # use base64 from dify 1.11
            'remember_me': remember_me,
            'language': language,
        })
        assert r == {"result": "success"}
        self.options['headers']['x-csrf-token'] = self.session.cookies.get("csrf_token")
        return self.session.cookies

    def set_tokens(self, *, csrf, access):
        """workaround for federated login"""
        self.session.cookies.set(name="__Host-csrf_token", value=csrf)
        self.session.cookies.set(name="__Host-access_token", value=access)


        self.options['headers']['x-csrf-token'] = csrf

    @property
    def me(self) -> dict:
        url = f"{self.base_url}/account/profile"
        return self.request(url, "GET")

    @property
    def workspace(self) -> dict:
        url = f"{self.base_url}/features"
        return self.request(url, "GET")


class ConsoleDerived(API):
    def __init__(self, context: ConsoleUser):
        super().__init__()
        self.base_url = context.base_url
        self.session.cookies = context.session.cookies
        self.options = context.options



from typing import Protocol, Any


class ClientProtocol(Protocol):
    api_key: str
    base_url: str


class ModelAware:
    def __init__(self):
        self.model: str | None = None


class SDKProtocol(Protocol):
    client: Any


class Connectable:
    def connect(self) -> bool: ...

    def close(self): ...

    def __enter__(self):
        assert self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from __future__ import annotations

from davidkhala.ai.agent.dify.api import API
from davidkhala.ai.agent.dify.api.knowledge.model import DocumentModel


class Chunk(API):
    def __init__(self, d: DocumentModel, segment_id: str):
        super().__init__(d.api_key, f"{d.base_url}/segments/{segment_id}")

    def get(self):
        r = self.request(self.base_url, "GET")
        assert r['doc_form']  # optional value text_model
        return r['data']



from __future__ import annotations

import os
from pathlib import Path
from typing import Iterable
from urllib.parse import urlparse

import requests

from davidkhala.ai.agent.dify.api import API, Iterator
from davidkhala.ai.agent.dify.api.knowledge.model import DatasetModel, DocumentModel


class Dataset(API):
    def __init__(self, api_key: str, base_url="https://api.dify.ai/v1"):
        super().__init__(api_key, f"{base_url}/datasets")

    def paginate_datasets(self, page=1, size=20):
        r = self.request(self.base_url, "GET", params={
            'page': page,
            'limit': size,
        })
        return r

    def list_datasets(self) -> Iterable[list[DatasetModel]]:
        return Iterator(self.paginate_datasets, None)

    @property
    def ids(self):
        for sub_list in self.list_datasets():
            for dataset in sub_list:
                yield dataset['id']

    class Instance(API):
        def __init__(self, d: Dataset, dataset_id: str):
            super().__init__(d.api_key, f"{d.base_url}/{dataset_id}")

        def get(self)-> DatasetModel:
            d = self.request(self.base_url, "GET")
            return DatasetModel.model_validate(d)

        def upload(self, filename, *, path=None, url=None, document_id=None):
            """
            don't work for .html
            work for .md
            """
            files = {}
            if path:
                with open(path, 'rb') as f:
                    content = f.read()
                if not filename:
                    filename = os.path.basename(path)
            elif url:
                r = requests.get(url)
                r.raise_for_status()
                if not filename:
                    parsed_url = urlparse(url)
                    filename = Path(parsed_url.path).name
                content = r.content
            files['file'] = (filename, content)
            if document_id:
                # don't work for html
                r = requests.post(f"{self.base_url}/documents/{document_id}/update-by-file", files=files,
                                  **self.options)
            else:
                r = requests.post(f"{self.base_url}/document/create-by-file", files=files, **self.options)
            r = self.on_response(r)
            return r['document']

        def paginate_documents(self, page=1, size=20):
            return self.request(f"{self.base_url}/documents", "GET", params={
                'page': page,
                'limit': size
            })

        def list_documents(self) -> Iterable[DocumentModel]:
            for document_batch in Iterator(self.paginate_documents, None):
                for document in document_batch:
                    yield DocumentModel(**document)

        def has_document(self, name) -> bool:
            return any(name == item['name'] for row in self.list_documents() for item in row)



from typing import Literal

from openai import OpenAI

from davidkhala.ai.model import SDKProtocol, Connectable
from davidkhala.ai.model.embed import EmbeddingAware
from davidkhala.ai.model.chat import on_response, ChatAware
from davidkhala.ai.model.garden import GardenAlike


class Client(ChatAware, EmbeddingAware, SDKProtocol, GardenAlike, Connectable):
    def __init__(self, client: OpenAI):
        super().__init__()
        self.client: OpenAI = client
        self.encoding_format: Literal["float", "base64"] = "float"

    def connect(self):
        try:
            type(self).models.fget(self)
            return True
        except:  # TODO make specific
            return False

    @property
    def models(self):
        return self.client.models.list()

    def encode(self, *_input: str) -> list[list[float]]:
        response = self.client.embeddings.create(
            model=self.model,
            input=list(_input),
            encoding_format=self.encoding_format
        )
        return [item.embedding for item in response.data]

    def chat(self, *user_prompt, **kwargs):

        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.messages_from(*user_prompt),
            n=self.n,
            **kwargs
        )

        return on_response(response, self.n)

    def close(self):
        self.client.close()
